/* id_fd.c: Interdata floppy disk simulator

   Copyright (c) 2001-2002, Robert M Supnik

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   ROBERT M SUPNIK BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of Robert M Supnik shall not
   be used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from Robert M Supnik.

   fd		M46-630 floppy disk

   A diskette consists of 77 tracks, each with 26 sectors of 128B.
   Tracks are numbered 0-76, sectors 1-26.
*/

#include "id_defs.h"

#define FD_NUMTR	77				/* tracks/disk */
#define FD_NUMSC	26				/* sectors/track */
#define FD_NUMBY	128				/* bytes/sector */
#define FD_NUMLRN	(FD_NUMTR * FD_NUMSC)		/* LRNs/disk */
#define FD_SIZE		(FD_NUMLRN * FD_NUMBY)		/* bytes/disk */
#define FD_NUMDR	4				/* drives/controller */
#define UNIT_V_WLK	(UNIT_V_UF)			/* write locked */
#define UNIT_WLK	(1u << UNIT_V_UF)
#define UNIT_WPRT	(UNIT_WLK | UNIT_RO)		/* write protect */
#define LRN		u3				/* last LRN */
#define FNC		u4				/* last function */
#define GET_DA(x)	(((x) - 1) * FD_NUMBY)
#define GET_TRK(x)	(((x) - 1) % FD_NUMSC)
#define GET_SEC(x)	((((x) - 1) / FD_NUMSC) + 1)
#define LRN_BOOT	5				/* boot block LRN */

/* Command byte */

#define CMD_V_UNIT	4				/* unit */
#define CMD_M_UNIT	0x3
#define GET_UNIT(x)	(((x) >> CMD_V_UNIT) & CMD_M_UNIT)
#define CMD_V_FNC	0				/* function */
#define CMD_M_FNC	0xF
#define GET_FNC(x)	(((x) >> CMD_V_FNC) & CMD_M_FNC)
#define  FNC_RD		1				/* read */
#define  FNC_WR		2				/* write */
#define  FNC_RDID	3				/* read ID */
#define  FNC_RSTA	4				/* read status */
#define  FNC_DEL	5				/* write deleted */
#define  FNC_BOOT	6				/* boot */
#define  FNC_STOP	7				/* stop */
#define  FNC_RESET	8				/* reset */
#define  FNC_FMT	9				/* format NI */

/* Status byte, * = dynamic */

#define STA_WRP		0x80				/* *write prot */
#define STA_DEF		0x40				/* def track NI */
#define STA_DEL		0x20				/* del record NI */
#define STA_ERR		0x10				/* error */
#define STA_IDL		0x02				/* idle */
#define STA_OFL		0x01				/* fault */
#define STA_MASK	(STA_DEF|STA_DEL|STA_ERR|STA_BSY|STA_IDL)
#define SET_EX		(STA_ERR)			/* set EX */

/* Extended status, 6 bytes, * = dynamic */

#define ES_SIZE		6
#define ES0_HCRC	0x80				/* ID CRC NI */
#define ES0_DCRC	0x40				/* data CRC NI */
#define ES0_LRN		0x20				/* illegal LRN */
#define ES0_WRP		0x10				/* *write prot */
#define ES0_ERR		0x08				/* error */
#define ES0_DEF		0x04				/* def trk NI */
#define ES0_DEL		0x02				/* del rec NI */
#define ES0_FLT		0x01				/* fault */
#define ES1_TK0		0x80				/* track 0 */
#define ES1_NRDY	0x40				/* not ready */
#define ES1_NOAM	0x20				/* no addr mk NI */
#define ES1_CMD		0x10				/* illegal cmd */
#define ES1_SKE		0x08				/* seek err NI */
#define ES1_UNS		0x04				/* unsafe NI */
#define ES1_UNIT	0x03				/* unit # */

/* Processing options for commands */

#define C_RD		0x1				/* cmd reads disk */
#define C_WD		0x2				/* cmd writes disk */

extern uint32 int_req[INTSZ], int_enb[INTSZ];

uint32 fd_sta = 0;					/* status */
uint32 fd_cmd = 0;					/* command */
uint32 fd_db = 0;					/* data buffer */
uint8 fdxb[FD_NUMBY] = { 0 };				/* sector buffer */
uint32 fd_bptr = 0;					/* buffer pointer */
uint32 fd_es[ES_SIZE] = { 0 };				/* ext status */
uint32 fd_lrn = 0;					/* log rec # */
uint32 fd_stopioe = 1;					/* stop on error */
uint32 fd_arm = 0;					/* intr arm */
int32 fd_ctime = 100;					/* command time */
int32 fd_stime = 10;					/* seek, per LRN */
int32 fd_xtime = 1;					/* tr set time */

static uint32 ctab[16] = {
	0, C_RD, C_WD, 0,				/* 0, rd, wr, 0 */
	0, C_WD, C_RD, 0,				/* 0, del, boot, 0 */
	0, 0, 0, 0,
	0, 0, 0, 0 };

DEVICE fd_dev;
uint32 fd (uint32 dev, uint32 op, uint32 dat);
t_stat fd_svc (UNIT *uptr);
t_stat fd_reset (DEVICE *dptr);
t_bool fd_dte (UNIT *uptr, t_bool wr);
t_bool fd_wbuf (UNIT *uptr);
void fd_done (uint32 nsta, uint32 nes0, uint32 nes1);
void sched_seek (UNIT *uptr, int32 newlrn);

/* FD data structures

   fd_dev	FD device descriptor
   fd_unit	FD unit list
   fd_reg	FD register list
   fd_mod	FD modifier list
*/

DIB fd_dib = { d_FD, -1, v_FD, NULL, &fd, NULL };

UNIT fd_unit[] = {
	{ UDATA (&fd_svc, UNIT_FIX+UNIT_ATTABLE+UNIT_DISABLE+
		UNIT_BUFABLE+UNIT_MUSTBUF, FD_SIZE) },
	{ UDATA (&fd_svc, UNIT_FIX+UNIT_ATTABLE+UNIT_DISABLE+
		UNIT_BUFABLE+UNIT_MUSTBUF, FD_SIZE) },
	{ UDATA (&fd_svc, UNIT_FIX+UNIT_ATTABLE+UNIT_DISABLE+
		UNIT_BUFABLE+UNIT_MUSTBUF, FD_SIZE) },
	{ UDATA (&fd_svc, UNIT_FIX+UNIT_ATTABLE+UNIT_DISABLE+
		UNIT_BUFABLE+UNIT_MUSTBUF, FD_SIZE) }  };

REG fd_reg[] = {
	{ HRDATA (CMD, fd_cmd, 8) },
	{ HRDATA (STA, fd_sta, 8) },
	{ HRDATA (BUF, fd_db, 8) },
	{ HRDATA (LRN, fd_lrn, 16) },
	{ BRDATA (ESTA, fd_es, 16, 8, ES_SIZE) },
	{ BRDATA (DBUF, fdxb, 16, 8, FD_NUMBY) },
	{ HRDATA (DBPTR, fd_bptr, 8) },
	{ FLDATA (IREQ, int_req[l_FD], i_FD) },
	{ FLDATA (IENB, int_enb[l_FD], i_FD) },
	{ FLDATA (IARM, fd_arm, 0) },
	{ DRDATA (CTIME, fd_ctime, 24), PV_LEFT },
	{ DRDATA (STIME, fd_stime, 24), PV_LEFT },
	{ DRDATA (XTIME, fd_xtime, 24), PV_LEFT },
	{ FLDATA (STOP_IOE, fd_stopioe, 0) },
	{ HRDATA (DEVNO, fd_dib.dno, 8), REG_HRO },
	{ NULL }  };

MTAB fd_mod[] = {
	{ UNIT_WLK, 0, "write enabled", "WRITEENABLED", NULL },
	{ UNIT_WLK, UNIT_WLK, "write locked", "LOCKED", NULL },
	{ MTAB_XTD|MTAB_VDV, 0, "DEVNO", "DEVNO",
		&set_dev, &show_dev, NULL },
	{ 0 }  };

DEVICE fd_dev = {
	"FD", fd_unit, fd_reg, fd_mod,
	FD_NUMDR, 16, 20, 1, 16, 8,
	NULL, NULL, &fd_reset,
	NULL, NULL, NULL,
	&fd_dib, DEV_DISABLE };

/* Floppy disk: IO routine */

uint32 fd (uint32 dev, uint32 op, uint32 dat)
{
int32 i, u, t, fnc;
UNIT *uptr;

fnc = GET_FNC (fd_cmd);					/* get fnc */
u = GET_UNIT (fd_cmd);					/* get unit */
uptr = fd_dev.units + u;
switch (op) {						/* case IO op */
case IO_ADR:						/* select */
	return BY;					/* byte only */
case IO_RD:						/* read */
	if (fd_sta & (STA_IDL | STA_BSY)) return fd_db;	/* idle, busy? */
	if (fd_bptr < FD_NUMBY) fd_db = fdxb[fd_bptr++];/* get byte */
	if (fd_bptr >= FD_NUMBY) {			/* buf end? */
	    if (ctab[fnc] & C_RD) {			/* disk read? */
		sched_seek (uptr, uptr->LRN + 1);	/* sched read */
		fd_sta = fd_sta | STA_BSY;  }		/* set busy */
	    else fd_bptr = 0;  }			/* just wrap */
	return fd_db;					/* return buf */
case IO_WD:						/* write */
	if (fd_sta & STA_IDL) {				/* idle? */
	    fd_lrn = ((fd_lrn << 8) | dat) & DMASK16;	/* insert byte */
	    break;  }
	if (fd_bptr < FD_NUMBY)				/* if room, */
		fdxb[fd_bptr++] = fd_db = dat;		/* store byte */
	if (fd_bptr >= FD_NUMBY) {			/* buf end? */
	    if (ctab[fnc] & C_WD) {			/* disk write? */
		sched_seek (uptr, uptr->LRN + 1);	/* sched write */
		fd_sta = fd_sta | STA_BSY;  }		/* set busy */
	    else fd_bptr = 0;  }			/* just wrap */
	break;
case IO_SS:						/* status */
	t = fd_sta & STA_MASK;				/* get status */
	if ((uptr->flags & UNIT_ATT) == 0) t = t | STA_DU;
	if (t & SET_EX) t = t | STA_EX;			/* test for ex */
	return t;
case IO_OC:						/* command */
	fd_cmd = dat;					/* save cmd */
	fd_arm = int_chg (v_FD, dat, fd_arm);		/* upd int ctrl */
	fnc = GET_FNC (dat);				/* get fnc */
	u = GET_UNIT (dat);				/* get unit */
	uptr = fd_dev.units + u;
	if (fnc == FNC_RESET) fd_reset (&fd_dev);	/* reset? */
	else if (fnc == FNC_STOP) {			/* stop? */
	    if (fd_sta & STA_IDL) break;		/* idle? yawn */
	    if ((ctab[fnc] & C_WD) && fd_bptr &&	/* wrt & data? */
		    fd_wbuf (uptr)) break;		/* error? */
	    fd_done (0, 0, 0);				/* nrml done */
	    break;  }
	else if (fd_sta & STA_IDL) {			/* must be idle */
	    fd_sta = STA_BSY;				/* busy, !idle */
	    fd_bptr = 0;				/* init buffer */
	    if (fnc != FNC_RSTA) {			/* !rd status */
		for (i = 0; i < ES_SIZE; i++) fd_es[i] = 0;
		fd_es[1] = u;  }			/* init ext sta */
	    if (fnc == FNC_BOOT) t = LRN_BOOT;		/* unless boot */
	    else t = fd_lrn;				/* prog LRN */
	    if (ctab[fnc] & C_RD) sched_seek (uptr, t);	/* seek now? */
	    else sim_activate (uptr, fd_ctime);		/* start cmd */
	    uptr->LRN = t;  }				/* save LRN */
	break;  }
return 0;
}

/* Unit service; the action to be taken depends on command */

t_stat fd_svc (UNIT *uptr)
{
int32 i;
t_addr da;

switch (uptr->FNC) {					/* case on fnc */

case FNC_BOOT:						/* boot, buf empty */
case FNC_RD:						/* read, buf empty */
	if (fd_dte (uptr, FALSE)) return SCPE_OK;	/* xfr error? */
	da = GET_DA (uptr->LRN);			/* get disk addr */
	for (i = 0; i < FD_NUMBY; i++)			/* get sector */
	    fdxb[i] = *(((uint8 *) uptr->filebuf) + da + i);
	fd_bptr = 0;					/* init buf */
	uptr->LRN = uptr->LRN + 1;			/* next block */
	break;

case FNC_WR: case FNC_DEL:				/* write block */
	if (fd_bptr) {					/* not first time? */
	    if (fd_wbuf (uptr)) return SCPE_OK;		/* dump buf, err? */
	    uptr->LRN = uptr->LRN + 1;  }		/* next block */
	fd_bptr = 0;					/* init buf */
	break;

case FNC_RSTA:						/* read status */
	if (uptr->flags & UNIT_WPRT)			/* wr protected? */
	    fd_es[0] = fd_es[0] | ES0_WRP;
	if ((uptr->LRN % FD_NUMSC) == 0)		/* on track 0? */
	    fd_es[1] = fd_es[1] | ES1_TK0;
	if (uptr->flags & UNIT_BUF) {			/* attached? */
	    fd_es[2] = (uptr->LRN >> 8) & DMASK8;	/* return LRN */
	    fd_es[3] = uptr->LRN & DMASK8;  }
	else {
	    fd_es[0] = fd_es[0] | ES0_FLT;		/* else err */
	    fd_es[1] = fd_es[1] | ES1_NRDY;  }
	for (i = 0; i < ES_SIZE; i++) fdxb[i] = fd_es[i];	/* copy to buf */
	for (i = ES_SIZE; i < FD_NUMBY; i++) fdxb[i] = 0;
	break;

case FNC_RDID:						/* read ID */
	if (fd_dte (uptr, FALSE)) return SCPE_OK;	/* xfr err? */
	for (i = 0; i < FD_NUMBY; i++) fdxb[i] = 0;	/* clr buf */
	fdxb[0] = GET_TRK (uptr->LRN);			/* store track */
	fdxb[2] = GET_SEC (uptr->LRN);			/* store sector */
	break;

case FNC_FMT:						/* format */
default:
	fd_done (STA_ERR, ES0_ERR, ES1_CMD);		/* ill cmd */
	break;  }

fd_sta = fd_sta & ~STA_BSY;				/* clear busy */
if (fd_arm) SET_INT (v_FD);				/* if armed, int */
return SCPE_OK;
}

/* Schedule seek */

void sched_seek (UNIT *uptr, int32 newlrn)
{
int32 diff = newlrn - uptr->LRN;			/* LRN diff */

if (diff < 0) diff = -diff;				/* ABS */
if (diff < 10) diff = 10;				/* MIN 10 */
sim_activate (uptr, diff * fd_stime);			/* schedule */
return;
}

/* Dump accumulated write buffer */

t_bool fd_wbuf (UNIT *uptr)
{
int32 i;
t_addr da;

if (fd_dte (uptr, TRUE)) return TRUE;			/* xfr error? */
for (i = fd_bptr; i < FD_NUMBY; i++) fdxb[i] = fd_db;	/* pad sector */
da = GET_DA (uptr->LRN);				/* get disk addr */
for (i = 0; i < FD_NUMBY; i++)				/* write */
	*(((uint8 *) uptr->filebuf) + da + i) = fdxb[i];
return FALSE;
}

/* Command complete */

void fd_done (uint32 nsta, uint32 nes0, uint32 nes1)
{
fd_sta = (fd_sta | STA_IDL | nsta) & ~STA_BSY;		/* set idle */
if (fd_arm) SET_INT (v_FD);				/* if armed, int */
fd_es[0] = fd_es[0] | nes0;				/* set ext state */
fd_es[1] = fd_es[1] | nes1;
return;
}

/* Test for data transfer error */

t_bool fd_dte (UNIT *uptr, t_bool wr)
{
if ((uptr->flags & UNIT_BUF) == 0) {			/* not attached? */
	fd_done (STA_ERR, ES0_ERR | ES0_FLT, ES1_NRDY);
	return TRUE;  }
if (wr && (uptr->flags & UNIT_WPRT)) {			/* wr protected? */
	fd_done (STA_ERR, ES0_ERR | ES0_WRP, 0);
	return TRUE;  }
if ((uptr->LRN == 0) || (uptr->LRN > FD_NUMLRN)) {	/* bad LRN? */
	fd_done (STA_ERR, ES0_ERR | ES0_LRN, 0);
	return TRUE;  }
return FALSE;
}

/* Reset routine */

t_stat fd_reset (DEVICE *dptr)
{
int32 i;
UNIT *uptr;

fd_sta = STA_IDL | STA_BSY;				/* idle, busy */
fd_cmd = 0;						/* clear state */
fd_db = 0;
fd_lrn = 0;
fd_bptr = 0;
for (i = 0; i < ES_SIZE; i++) fd_es[i] = 0;		/* clr ext sta */
for (i = 0; i < FD_NUMBY; i++) fdxb[i] = 0;		/* clr xfr buf */
for (i = 0; i < FD_NUMDR; i++) {			/* loop thru units */
	uptr = fd_dev.units + i;
	sim_cancel (uptr);				/* stop drive */
	uptr->LRN = 0;					/* clear state */
	uptr->FNC = 0;  }
CLR_INT (v_FD);						/* clear int */
CLR_ENB (v_FD);						/* disable int */
fd_arm = 0;						/* disarm int */
return SCPE_OK;
}
